==> conson_xp-0.11.1.tar.gz: conson_xp-0.11.1/src/xp/services/conbus/conbus_service.py <==
"""Conbus Client Send Service for TCP communication with Conbus servers.

This service implements a TCP client that connects to Conbus servers and sends
various types of telegrams including discover, version, and sensor data requests.
"""

import logging
import socket
from datetime import datetime
from typing import Any, List, Optional

from typing_extensions import Callable

from xp.models import (
    ConbusConnectionStatus,
    ConbusRequest,
    ConbusResponse,
)
from xp.models.conbus.conbus_client_config import ConbusClientConfig
from xp.models.response import Response
from xp.models.telegram.system_function import SystemFunction
from xp.services.conbus.conbus_connection_pool import ConbusConnectionPool
from xp.utils.checksum import calculate_checksum


class ConbusError(Exception):
    """Raised when Conbus client send operations fail"""

    pass


class ConbusService:
    """
    TCP client service for sending telegrams to Conbus servers.

    Manages TCP socket connections, handles telegram generation and transmission,
    and processes server responses.
    """

    def __init__(
        self,
        client_config: ConbusClientConfig,
    ):
        """Initialize the Conbus client send service

        Args:
            client_config: Optional ConbusClientConfig for dependency injection
        """
        self.client_config: ConbusClientConfig = client_config or ConbusClientConfig()
        self.is_connected = False
        self.last_activity: Optional[datetime] = None

        # Set up logging
        self.logger = logging.getLogger(__name__)

        # Initialize connection pool
        self._connection_pool = ConbusConnectionPool.get_instance()
        self._connection_pool.initialize(self.client_config)

    def get_config(self) -> ConbusClientConfig:
        """Get current client configuration"""
        return self.client_config

    def connect(self) -> Response:
        """Test connection using the connection pool"""
        try:
            # Test connection by acquiring and immediately releasing
            with self._connection_pool:
                self.is_connected = True
                self.last_activity = datetime.now()

                self.logger.info(
                    f"Connection pool ready for {self.client_config.conbus.ip}:{self.client_config.conbus.port}"
                )

                return Response(
                    success=True,
                    data={
                        "message": f"Connection pool ready for {self.client_config.conbus.ip}:{self.client_config.conbus.port}",
                    },
                    error=None,
                )

        except Exception as e:
            error_msg = f"Failed to establish connection pool to {self.client_config.conbus.ip}:{self.client_config.conbus.port}: {e}"
            self.logger.error(error_msg)
            self.is_connected = False
            return Response(success=False, data=None, error=error_msg)

    def disconnect(self) -> None:
        """Close connection pool (graceful shutdown)"""
        try:
            self._connection_pool.close()
            self.logger.info("Connection pool closed")
        except Exception as e:
            self.logger.error(f"Error closing connection pool: {e}")
        finally:
            self.is_connected = False

    def get_connection_status(self) -> ConbusConnectionStatus:
        """Get current connection status"""
        return ConbusConnectionStatus(
            connected=self.is_connected,
            ip=self.client_config.conbus.ip,
            port=self.client_config.conbus.port,
            last_activity=self.last_activity,
        )

    @staticmethod
    def _parse_telegrams(raw_data: str) -> List[str]:
        """Parse raw data and extract telegrams using < and > delimiters"""
        telegrams: list[str] = []
        if not raw_data:
            return telegrams

        # Find all telegram patterns <...>
        start_pos = 0
        while True:
            # Find the start of next telegram
            start_idx = raw_data.find("<", start_pos)
            if start_idx == -1:
                break

            # Find the end of this telegram
            end_idx = raw_data.find(">", start_idx)
            if end_idx == -1:
                # Incomplete telegram at the end
                break

            # Extract telegram including < and >
            telegram = raw_data[start_idx : end_idx + 1]
            if telegram.strip():
                telegrams.append(telegram.strip())

            start_pos = end_idx + 1

        return telegrams

    def receive_responses(
        self,
        timeout: float = 0.1,
        receive_callback: Optional[Callable[[list[str]], None]] = None,
    ) -> List[str]:
        """Receive responses from the server and properly split telegrams"""
        import time

        all_telegrams = []
        start_time = time.time()

        try:
            with self._connection_pool as connection:
                while time.time() - start_time < timeout:
                    # Call _receive_responses_with_connection with a short timeout for each iteration
                    telegrams = self._receive_responses_with_connection(
                        connection, timeout=0.1
                    )
                    if receive_callback is not None:
                        receive_callback(telegrams)
                    all_telegrams.extend(telegrams)

                    # Small sleep to avoid busy waiting when no data
                    time.sleep(0.01)
        except Exception as e:
            self.logger.error(f"Error receiving responses: {e}")
            return []

        return all_telegrams

    def _receive_responses_with_connection(
        self,
        connection: socket.socket,
        timeout: float = 1.0,
        receive_callback: Optional[Callable[[list[str]], None]] = None,
    ) -> List[str]:
        """Receive responses from the server using a specific connection"""
        accumulated_data = ""

        try:
            # Set a shorter timeout for receiving responses
            original_timeout = connection.gettimeout()
            connection.settimeout(timeout)  # 1 second timeout for responses

            while True:
                try:
                    data = connection.recv(1024)
                    if not data:
                        break

                    # Accumulate all received data
                    message = data.decode("latin-1")
                    if receive_callback is not None:
                        parsed_telegrams = self._parse_telegrams(message)
                        receive_callback(parsed_telegrams)
                    accumulated_data += message
                    self.last_activity = datetime.now()
                    connection.settimeout(0.1)  # 2 second timeout for responses

                except socket.timeout:
                    # No more data available
                    break

            # Restore original timeout
            connection.settimeout(original_timeout)

        except Exception as e:
            self.logger.error(f"Error receiving responses: {e}")

        # Parse telegrams from accumulated data
        telegrams = self._parse_telegrams(accumulated_data)
        for telegram in telegrams:
            self.logger.info(f"Received telegram: {telegram}")

        return telegrams

    def send_telegram(
        self,
        serial_number: str,
        system_function: SystemFunction,
        data: str,
        receive_callback: Optional[Callable[[list[str]], None]] = None,
    ) -> ConbusResponse:
        """Send custom telegram with specified function and data point codes"""
        # Generate custom system telegram: <S{serial}F{function}{data_point}{checksum}>
        function_code = system_function.value
        telegram_body = f"S{serial_number}F{function_code}D{data}"
        checksum = calculate_checksum(telegram_body)
        telegram = f"<{telegram_body}{checksum}>"

        return self.send_raw_telegram(telegram, receive_callback)

    def send_telegram_body(
        self,
        telegram_body: str,
        receive_callback: Optional[Callable[[list[str]], None]] = None,
    ) -> ConbusResponse:
        """Send custom telegram with specified function and data point codes"""
        checksum = calculate_checksum(telegram_body)
        telegram = f"<{telegram_body}{checksum}>"

        return self.send_raw_telegram(telegram, receive_callback)

    def send_raw_telegram(
        self,
        telegram: Optional[str] = None,
        receive_callback: Optional[Callable[[list[str]], None]] = None,
    ) -> ConbusResponse:
        """Send telegram using connection pool with automatic acquire/release"""
        request = ConbusRequest(telegram=telegram)

        try:
            # Use context manager for automatic connection management
            with self._connection_pool as connection:

                # Draining event waiting to be read (wait time 0)
                responses = self._receive_responses_with_connection(connection, 0.001)
                self.logger.info(f"Purged telegram: {responses}")

                # Send telegram
                if telegram is not None:
                    connection.send(telegram.encode("latin-1"))
                    self.logger.info(f"Sent telegram: {telegram}")

                self.last_activity = datetime.now()
                self.is_connected = True  # Update connection status

                # Receive responses
                responses = self._receive_responses_with_connection(
                    connection, 0.1, receive_callback
                )

                return ConbusResponse(
                    success=True,
                    request=request,
                    sent_telegram=telegram,
                    received_telegrams=responses,
                )
                # Connection automatically released here

        except Exception as e:
            error_msg = f"Failed to send telegram: {e}"
            self.logger.error(error_msg)
            self.is_connected = False  # Update connection status on error
            return ConbusResponse(
                success=False,
                request=request,
                error=error_msg,
            )

    def send_raw_telegrams(self, telegrams: List[str]) -> ConbusResponse:
        self.logger.info(f"send_raw_telegrams: {telegrams}")
        all_telegrams = "".join(telegrams)
        return self.send_raw_telegram(all_telegrams)

    def __enter__(self) -> "ConbusService":
        """Context manager entry"""
        return self

    def __exit__(
        self,
        _exc_type: Optional[type],
        _exc_val: Optional[Exception],
        _exc_tb: Optional[Any],
    ) -> None:
        """Context manager exit - ensure connection is closed"""
        self.disconnect()

==> conson_xp-0.11.3.tar.gz: conson_xp-0.11.3/src/xp/services/conbus/conbus_service.py <==
"""Conbus Client Send Service for TCP communication with Conbus servers.

This service implements a TCP client that connects to Conbus servers and sends
various types of telegrams including discover, version, and sensor data requests.
"""

import logging
import socket
from datetime import datetime
from typing import Any, List, Optional

from typing_extensions import Callable

from xp.models import (
    ConbusConnectionStatus,
    ConbusRequest,
    ConbusResponse,
)
from xp.models.conbus.conbus_client_config import ConbusClientConfig
from xp.models.response import Response
from xp.models.telegram.system_function import SystemFunction
from xp.services.conbus.conbus_connection_pool import ConbusConnectionPool
from xp.utils.checksum import calculate_checksum


class ConbusError(Exception):
    """Raised when Conbus client send operations fail"""

    pass


class ConbusService:
    """
    TCP client service for sending telegrams to Conbus servers.

    Manages TCP socket connections, handles telegram generation and transmission,
    and processes server responses.
    """

    def __init__(
        self,
        client_config: ConbusClientConfig,
        connection_pool: ConbusConnectionPool,
    ):
        """Initialize the Conbus client send service

        Args:
            client_config: ConbusClientConfig for dependency injection
            connection_pool: ConbusConnectionPool for dependency injection
        """
        self.client_config: ConbusClientConfig = client_config
        self.is_connected = False
        self.last_activity: Optional[datetime] = None

        # Set up logging
        self.logger = logging.getLogger(__name__)

        # Use injected connection pool
        self._connection_pool = connection_pool

    def get_config(self) -> ConbusClientConfig:
        """Get current client configuration"""
        return self.client_config

    def connect(self) -> Response:
        """Test connection using the connection pool"""
        try:
            # Test connection by acquiring and immediately releasing
            with self._connection_pool:
                self.is_connected = True
                self.last_activity = datetime.now()

                self.logger.info(
                    f"Connection pool ready for {self.client_config.conbus.ip}:{self.client_config.conbus.port}"
                )

                return Response(
                    success=True,
                    data={
                        "message": f"Connection pool ready for {self.client_config.conbus.ip}:{self.client_config.conbus.port}",
                    },
                    error=None,
                )

        except Exception as e:
            error_msg = f"Failed to establish connection pool to {self.client_config.conbus.ip}:{self.client_config.conbus.port}: {e}"
            self.logger.error(error_msg)
            self.is_connected = False
            return Response(success=False, data=None, error=error_msg)

    def disconnect(self) -> None:
        """Close connection pool (graceful shutdown)"""
        try:
            self._connection_pool.close()
            self.logger.info("Connection pool closed")
        except Exception as e:
            self.logger.error(f"Error closing connection pool: {e}")
        finally:
            self.is_connected = False

    def get_connection_status(self) -> ConbusConnectionStatus:
        """Get current connection status"""
        return ConbusConnectionStatus(
            connected=self.is_connected,
            ip=self.client_config.conbus.ip,
            port=self.client_config.conbus.port,
            last_activity=self.last_activity,
        )

    @staticmethod
    def _parse_telegrams(raw_data: str) -> List[str]:
        """Parse raw data and extract telegrams using < and > delimiters"""
        telegrams: list[str] = []
        if not raw_data:
            return telegrams

        # Find all telegram patterns <...>
        start_pos = 0
        while True:
            # Find the start of next telegram
            start_idx = raw_data.find("<", start_pos)
            if start_idx == -1:
                break

            # Find the end of this telegram
            end_idx = raw_data.find(">", start_idx)
            if end_idx == -1:
                # Incomplete telegram at the end
                break

            # Extract telegram including < and >
            telegram = raw_data[start_idx : end_idx + 1]
            if telegram.strip():
                telegrams.append(telegram.strip())

            start_pos = end_idx + 1

        return telegrams

    def receive_responses(
        self,
        timeout: float = 0.1,
        receive_callback: Optional[Callable[[list[str]], None]] = None,
    ) -> List[str]:
        """Receive responses from the server and properly split telegrams"""
        import time

        all_telegrams = []
        start_time = time.time()

        try:
            with self._connection_pool as connection:
                while time.time() - start_time < timeout:
                    # Call _receive_responses_with_connection with a short timeout for each iteration
                    telegrams = self._receive_responses_with_connection(
                        connection, timeout=0.1
                    )
                    if receive_callback is not None:
                        receive_callback(telegrams)
                    all_telegrams.extend(telegrams)

                    # Small sleep to avoid busy waiting when no data
                    time.sleep(0.01)
        except Exception as e:
            self.logger.error(f"Error receiving responses: {e}")
            return []

        return all_telegrams

    def _receive_responses_with_connection(
        self,
        connection: socket.socket,
        timeout: float = 1.0,
        receive_callback: Optional[Callable[[list[str]], None]] = None,
    ) -> List[str]:
        """Receive responses from the server using a specific connection"""
        accumulated_data = ""

        try:
            # Set a shorter timeout for receiving responses
            original_timeout = connection.gettimeout()
            connection.settimeout(timeout)  # 1 second timeout for responses

            while True:
                try:
                    data = connection.recv(1024)
                    if not data:
                        break

                    # Accumulate all received data
                    message = data.decode("latin-1")
                    if receive_callback is not None:
                        parsed_telegrams = self._parse_telegrams(message)
                        receive_callback(parsed_telegrams)
                    accumulated_data += message
                    self.last_activity = datetime.now()
                    connection.settimeout(0.1)  # 2 second timeout for responses

                except socket.timeout:
                    # No more data available
                    break

            # Restore original timeout
            connection.settimeout(original_timeout)

        except Exception as e:
            self.logger.error(f"Error receiving responses: {e}")

        # Parse telegrams from accumulated data
        telegrams = self._parse_telegrams(accumulated_data)
        for telegram in telegrams:
            self.logger.info(f"Received telegram: {telegram}")

        return telegrams

    def send_telegram(
        self,
        serial_number: str,
        system_function: SystemFunction,
        data: str,
        receive_callback: Optional[Callable[[list[str]], None]] = None,
    ) -> ConbusResponse:
        """Send custom telegram with specified function and data point codes"""
        # Generate custom system telegram: <S{serial}F{function}{data_point}{checksum}>
        function_code = system_function.value
        telegram_body = f"S{serial_number}F{function_code}D{data}"
        checksum = calculate_checksum(telegram_body)
        telegram = f"<{telegram_body}{checksum}>"

        return self.send_raw_telegram(telegram, receive_callback)

    def send_telegram_body(
        self,
        telegram_body: str,
        receive_callback: Optional[Callable[[list[str]], None]] = None,
    ) -> ConbusResponse:
        """Send custom telegram with specified function and data point codes"""
        checksum = calculate_checksum(telegram_body)
        telegram = f"<{telegram_body}{checksum}>"

        return self.send_raw_telegram(telegram, receive_callback)

    def send_raw_telegram(
        self,
        telegram: Optional[str] = None,
        receive_callback: Optional[Callable[[list[str]], None]] = None,
    ) -> ConbusResponse:
        """Send telegram using connection pool with automatic acquire/release"""
        request = ConbusRequest(telegram=telegram)

        try:
            # Use context manager for automatic connection management
            with self._connection_pool as connection:

                # Draining event waiting to be read (wait time 0)
                responses = self._receive_responses_with_connection(connection, 0.001)
                self.logger.info(f"Purged telegram: {responses}")

                # Send telegram
                if telegram is not None:
                    connection.send(telegram.encode("latin-1"))
                    self.logger.info(f"Sent telegram: {telegram}")

                self.last_activity = datetime.now()
                self.is_connected = True  # Update connection status

                # Receive responses
                responses = self._receive_responses_with_connection(
                    connection, 0.1, receive_callback
                )

                return ConbusResponse(
                    success=True,
                    request=request,
                    sent_telegram=telegram,
                    received_telegrams=responses,
                )
                # Connection automatically released here

        except Exception as e:
            error_msg = f"Failed to send telegram: {e}"
            self.logger.error(error_msg)
            self.is_connected = False  # Update connection status on error
            return ConbusResponse(
                success=False,
                request=request,
                error=error_msg,
            )

    def send_raw_telegrams(self, telegrams: List[str]) -> ConbusResponse:
        self.logger.info(f"send_raw_telegrams: {telegrams}")
        all_telegrams = "".join(telegrams)
        return self.send_raw_telegram(all_telegrams)

    def __enter__(self) -> "ConbusService":
        """Context manager entry"""
        return self

    def __exit__(
        self,
        _exc_type: Optional[type],
        _exc_val: Optional[Exception],
        _exc_tb: Optional[Any],
    ) -> None:
        """Context manager exit - ensure connection is closed"""
        self.disconnect()
